
%initial code by Sigurd Fjerdingen. 
%Finding activation function output ratio and the methods remove_neuron, 
%pruneO and pruneW are new for this report

classdef RBFN < MRL.FA
% Implementation of a static three-layer Radial Basis Function network

    properties (SetAccess = private)
        nodeCount = 1;
        
        mu % center vector
        sigma % width vector
        w % weight vector for approximation
        k_g = 1/sqrt(pi);
        alpha = 1; % Training rate. Determines the step-size in the gradient direction for gradient descent.
        
        stateMin
        stateMax
    end % properties
    
    methods
        
        function obj = RBFN(indim, outdim, varargin)
            % RBFN(indim, outdim, min1, max1, step1, min2, max2, step2, ...)
            % 
            % Initializes the RBFN by distributing the RBFs such that the centers are
            % equally spaced between min and max, and the widths are equal to the
            % distance between the centers
            
            optargin = size(varargin,2);
            if optargin ~= 3*indim,
                error('Input dimension does not match number of parameters');
            end
            
            if indim <= 0 || outdim <= 0,
                error('indim and outdim must be positive integers');
            end
            
            sMin = zeros(indim,1); sMax = sMin; sStep = sMin;
            for i = 1:indim,
                sStep(i) = varargin{3*i - 0};
                sMin(i) = varargin{3*i - 2};
                sMax(i) = varargin{3*i - 1};
            end
            
            obj.indim = indim;
            obj.outdim = outdim;
            obj.stateMin = sMin;
            obj.stateMax = sMax;
            
            no_hidden = zeros(indim,1);
            for i = 1:indim,
                % Number of hidden nodes per dimension is min->max times step_size
                no_hidden(i) = length(sMin(i):sStep(i):sMax(i));
                if no_hidden(i) == 0,
                    error(['Dimension discretization for RBF dimension variable ' num2str(i) ...
                        ' is too large in comparison to the min/max span']);
                end
                obj.nodeCount = obj.nodeCount * no_hidden(i);
            end
            
            % Initialize mu, sigma, w
            obj.mu = zeros(obj.nodeCount,indim);
            obj.sigma = obj.mu;
            obj.w = zeros(obj.nodeCount,outdim);
            
            sub_cell = cell(1,length(no_hidden));
            for i = 1:obj.nodeCount,
                [sub_cell{:}] = ind2sub(no_hidden', i);
                for dim = 1:indim,
                    obj.mu(i,dim) = sMin(dim) + sStep(dim) * (sub_cell{dim}-1);
                    
                    obj.sigma(i,dim) = sStep(dim);
                end
            end
        end % function RBFN
        
        
        
        function [val act_outp] = Calculate(obj,inputs) 
            % Calculate output of RBFN based on given input vector.
            % Input vector must be a vector of length equal to the RBFN input dimension.
            
            if length(inputs) ~= obj.indim,
                error('Input vector dimension does not match RBFN input dimension');
            end
            
            states_expanded = repmat(inputs,obj.nodeCount,1);
            product = ((states_expanded - obj.mu) ./ (obj.sigma)).^2;
            g = obj.k_g * exp(-sum(product,2));%output of hidden layer
            
            g_expanded = repmat(g,1,obj.outdim);
            val = sum(obj.w .* g_expanded);
            
            trust = sum(g); 
            
            %finding activation function output ratio for each neuron for the given inputs
            outpMax = max(g);
             act_outp = zeros(length(g),1);
             for i = 1 : length(g)
                act_outp(i) = abs( g(i) / outpMax );
             end
                  
        end % function Calculate
        
        
        
        function obj = Train(obj,inputs,diff)
            % Trains the weight of the network according to the objective function
            % K(w) = 1/2 * (y - RBFN(inputs,w))^2
            % using gradient descent.
            % diff = y - RBFN(inputs,w), i.e. difference between observed and calculated output
            %        for the current inputs.
            
            if length(inputs) ~= obj.indim,
                error('Input vector dimension does not match RBFN input dimension');
            end
            
            % Calculate derivatives of RBFN
            states_expanded = repmat(inputs,obj.nodeCount,1);
            product = ((states_expanded - obj.mu) ./ (obj.sigma)).^2;
            dfdw = obj.k_g * exp(-sum(product,2));
            
            % Adjust weights according to gradient descent
            obj.w = obj.w + obj.alpha .* repmat(diff,obj.nodeCount,1) .* repmat(dfdw,1,obj.outdim);
        end % function Derivative
        
        
        
        function obj = SetWeights(obj,weights)
            % Set the internal weights of the network to the given weight vector. The vector
            % should have dimensionality equal to the RBFN output dimension.
            
            if length(weights) ~= obj.outdim,
                error('Weight vector dimension does not match RBFN output dimension');
            end
           
            obj.w = ndgrid(weights,ones(obj.nodeCount,1))';
        end % function SetWeights
        
        
        function obj = SetTrainingRate(obj,alpha)
            if alpha > 1 || alpha < 0,
                error('The training rate should be in the interval [0,1]');
            end
            obj.alpha = alpha;
        end % function SetTrainingRate
        
        
      %  function Plot(obj, outdim, resolution)
      %      if outdim <= 0 || outdim > obj.outdim,
      %          error('Erroneous output dimension selection.');
      %      end
            
      %      if obj.indim > 2,
      %          error('Plotting only supported for 1 and 2 dimensional input');
      %      end
            
       %     x = obj.stateMin(1):resolution:obj.stateMax(1);
       %     if obj.indim == 1,
       %         f = zeros(1,length(x));
       %         for i = 1:length(x),
       %             f(i) = obj.Calculate(x(i));
       %         end
       %         plot(x,f);
       %     elseif obj.indim == 2,
       %         y = obj.stateMin(2):resolution:obj.stateMax(2);
       %         f = zeros(length(y),length(x));
       %         for yi = 1:length(y),
       %             for xi = 1:length(x),
       %                 f(yi,xi) = obj.Calculate([x(xi) y(yi)]);
       %             end
       %         end
       %         [yg,xg] = meshgrid(x,y);
       %         surf(xg,yg,f);
       %     end
       % end % function Plot
        
        function obj = remove_neuron(obj, index, n_inp)
                      
            %disp(['in remove neuron, index is  ' num2str(neuron) ]);
            
            obj.nodeCount = obj.nodeCount - 1;
            
            %disp(['in pruneO, nodeCount is now ' num2str(obj.nodeCount) ]);
            weight = obj.w;
            length_w = length(weight);
            new_weight = weight([1:index-1 index+1:length_w]');
            obj.w = new_weight(:);
                
            %disp(['in pruneO, weights are now ' num2str(length(obj.w)) ]);
               
            if n_inp == 1
                centers = obj.mu;
                length_mu = length(centers);
                centers_new = centers([1:index-1 index+1:length_mu]',1);
                %centers_new2 = centers([1:index-1 index+1:length_mu]',2);
                obj.mu = centers_new;
           elseif n_inp == 2
                centers = obj.mu;
                length_mu = length(centers);
                centers_new1 = centers([1:index-1 index+1:length_mu]',1);
                centers_new2 = centers([1:index-1 index+1:length_mu]',2);
                obj.mu = [centers_new1 centers_new2];    
           elseif n_inp == 3
                centers = obj.mu;
                length_mu = length(centers);
                centers_new1 = centers([1:index-1 index+1:length_mu]',1);
                centers_new2 = centers([1:index-1 index+1:length_mu]',2);
                centers_new3 = centers([1:index-1 index+1:length_mu]',3);
                obj.mu = [centers_new1 centers_new2 centers_new3];                 
           end
                
           %disp(['in pruneO, number centers are now ' num2str(length(obj.mu)) ]);
                
           if n_inp == 1
                width = obj.sigma;
                length_sigma = length(width);
                sigma_new = width([1:index-1 index+1:length_sigma]');
                obj.sigma = sigma_new;
           elseif n_inp ==2
                width = obj.sigma;
                length_sigma = length(width);
                sigma_new1 = width([1:index-1 index+1:length_sigma]',1);
                sigma_new2 = width([1:index-1 index+1:length_sigma]',2);
                obj.sigma = [sigma_new1 sigma_new2];       
           elseif n_inp == 3
                width = obj.sigma;
                length_sigma = length(width);
                sigma_new1 = width([1:index-1 index+1:length_sigma]',1);
                sigma_new2 = width([1:index-1 index+1:length_sigma]',2);
                sigma_new3 = width([1:index-1 index+1:length_sigma]',3);
                obj.sigma = [sigma_new1 sigma_new2 sigma_new3];
           end
           %disp(['after removing- centers:  ' num2str(length(obj.mu)) ' weights: ' num2str(length(obj.w)) ' sigmas: ' num2str(length(obj.sigma)) ]);
           %disp(['removed neuron ' num2str(neuron) ' and nodeCount is ' num2str(obj.nodeCount)]);
        
        end %remove_neuron
        
        
        function [obj act_outp pruned_units prun_Out] = pruneO(obj, act_outp, outp_th, n_inp) 
            %prunes based on size of hidden neuron output
            %Returns the obj, normalized activation function output, a list of
            %which units that have been pruned and the current pruning
            %threshold.
            %Takes as input the a list of previous normalized activation function output,
            %pruning percent and number of NN inputs

            iterations = length(obj.mu);
            
           if iterations > 2
                m = 0;
                small_output = zeros(iterations,1);
                prun_Out = outp_th * max(act_outp); 

                %finds the neurons that are to be pruned away
                for i = 1 : iterations
                    if act_outp(i) < prun_Out
                        m = m+1;
                        small_output(m) = i;
                    end
                end 

                if m == 0
                    pruned_units = 0;
                else
                    pruned_units = zeros(m,1);

                    %disp(['in pruneO ' num2str(m) ' units to be removed']);
                    %prunes the network
                    for k = 1 : m 
                        if k == 1
                            index = small_output(k);
                            pruned_units(k) = small_output(k);
                        end
                        if k > 1
                            index = small_output(k) - (k - 1);
                            pruned_units(k) = small_output(k);
                        end               

                        %disp(['in pruneO, index is ' num2str(index) ]);
                        obj = obj.remove_neuron(index, n_inp);    
                        act_len = length(act_outp);
                        act_tmp = act_outp([1:index-1 index+1:act_len]');
                        act_outp = act_tmp;
                        %disp(['in pruneO, activation func are now ' num2str(length(obj.sigma)) ]);
                    end %pruning
                end  %if else
            
            else
                prun_Out = 0;
                pruned_units = 0;
            end
               
                
            %disp(['done pruneO, number centers are now ' num2str(length(obj.mu)) ]);
        end %function pruneO
        
        
        function [obj n_prunTimes pruned_units min_size act_outp] = pruneW(obj, weight_th, n_inp, n_prunTimes, act_outp)
            %prunes based on the magnitude of the weights.
            %Returns the obj, the number the NN has been pruned, a list of
            %which units that have been pruned and the current pruning
            %threshold.
            %Takes as input the pruning percent, number of NN inputs,
            %number of times the NN has been pruned before.
            
            %disp(['in pruneW, size obj ' num2str(length(obj.mu))]);
            
            weight = obj.w;
            
            if length(weight) > 2
                iterations = length(weight);
                j = 0;
                euc_dist = norm(weight,2);
                if n_prunTimes <=3 %to avoid the threshold to grow to infinity
                    min_size = (euc_dist * weight_th * n_prunTimes); %threshold for size of weights to be removed 
                else
                    min_size = (euc_dist * weight_th *1);
                   % disp(['in RBFN.m, this network has been pruned more than three times']);
                end
                small_weights = zeros(iterations,1);

                %finds the weights that are to be pruned away
                for i = 1 : iterations
                    if abs(weight(i)) < min_size
                        j = j + 1;
                        small_weights(j) = i;                        
                    end
                end

                if j == 0
                    pruned_units = 0;
                else
                    pruned_units = zeros(j,1);
                end
                %disp(['in pruneW, ' num2str(j) ' weights/neurons to be removed' ]);

                for k = 1 : j
                    if k == 1
                        index = small_weights(k);
                        pruned_units(k) = small_weights(k); 
                    end
                    if k > 1
                        index = small_weights(k) - (k - 1);
                        pruned_units(k) = small_weights(k);
                    end                     
                    obj = obj.remove_neuron(index, n_inp);

                   %if the network also is pruned by node output
                    act_len = length(act_outp);
                    act_tmp = act_outp([1:index-1 index+1:act_len]');
                    act_outp = act_tmp;
                end

                if j > 0
                    n_prunTimes = n_prunTimes + 1;
                end
                
            else
               % disp(['this network can not be pruned anymore']);
                pruned_units = 0;
                min_size = 0;
                %act_outp = act_outp;
                %n_prunTimes = n_prunTimes;
            end
      
            %disp(['in pruneW, done' ]);
        end %function pruneW
        
        
    end % methods
    
end % classdef