
%----------------------------code by Sigurd Fjerdingen--------------------
classdef ARBFN < MRL.FA
% Implementation of a dynamic (additive) three-layer Radial Basis Function network
% using the MRAN algorithm.
% 
% 
% Tuning parameters for the MRAN algorithm adding strategy:
% 
% - epsilon_min: Euclidean distance between current (input) state and state
% (center) of the nearest vector stored in memory. This distance must be
% larger than "epsilon_min" in order for new vectors to be added to memory.
% 
% - e_min: Value difference between calculated output from memory and
% actual (wanted) output. This error must be larger than "e_min" in order
% for new vectors to be added to memory.
% 
% - e_minl: Based on a window of size "M_window" (which also may be set)
% the root mean square of all the last M incoming output differences (errors)
% are calculated. Must be larger than "e_minl" in order for new vectors to be
% added to memory. This means that the mean error needs to be significant
% over time in order for new vectors to be added (low-pass filter).


	properties (SetAccess = private)
		nodeCount = 0;
		
		mu = []; % center vector
		sigma = []; % width vector (this may make the RBFs elliptical?)
        sigma_0 = 1; % width of the first node in the network (when there is nothing to compare against)
		w = []; % weight vector
        w_0 = 0; % bias ( f(x) = w_0 + RBFN(...) )
        
        Kappa = 0.5; % Controls overlap between adjacent RBFs when adding a new RBF
        epsilon_n = 1; % Controls minimum distance to add new RBF
        e_min = 1; % Controls minimum approximation error to add new RBF
        e_minl = 1; % Threshold for filtering out add<->prune looping on noisy data
        
        M_window = 50; % Max window size when calculating the add<->prune filtering
        M_window_current = 0; % Current window size (important when too few measurements to max out window)
        e_i = [];
        
        P_0 = 1; % Initial value for the Kalman filter error covariance.
        P_n = []; % Kalman filter error covariance matrix.
        R_n = []; % Variance of Kalman filter measurement noise. Output vector dimensionality.
        Q = 0.4; % Kalman filter added random walk noise (to avoid too rapid convergence).
	end % properties (SetAccess = private)
	
	
	
	methods
        
        function res = GetNodeCount(obj)
            res = obj.nodeCount;
        end % function GetNodeCount
		
		function obj = ARBFN(no_indim, no_outdim, epsilon_max, epsilon_min, e_min, e_minl)
            if no_indim <= 0 || no_outdim <= 0,
				error('Input and output dimensionality must be positive integers.');
            end
            if no_outdim > 1,
                error('Only 1-dimensional output supported right now.');
            end
			obj.indim = floor(no_indim);
			obj.outdim = floor(no_outdim);
            
            obj.epsilon_n = epsilon_min;
            obj.e_min = e_min;
            obj.e_minl = e_minl;
            
            obj.e_i = zeros(obj.M_window, 1);
            
            obj.R_n = obj.P_0 * eye(obj.outdim);
		end % function ARBFN
        
        
        function obj = SetProperties(obj, varargin)
            for i = 1:2:size(varargin,2), % iterate through argument list
                if ~isa(varargin{i},'char'),
                    error(['Erroneous argument type (argument ' num2str(i) '): ' class(varargin{i})]);
                end
                
                switch varargin{i}
                    case 'Kappa'
                        obj.Kappa = varargin{i+1};
                    case 'epsilon_n'
                        obj.epsilon_n = varargin{i+1};
                        disp();
                    case 'e_min'
                        obj.e_min = varargin{i+1};
                    case 'e_minl'
                        obj.e_minl = varargin{i+1};
                    case 'P_0'
                        obj.P_0 = varargin{i+1};
                    case 'Q'
                        obj.Q = varargin{i+1};
                    otherwise
                        disp(['Unknown parameter: ' varargin{i}])
                end
            end
        end % function SetProperties
        
        
        % RBF core calculation
        function g = Rbf(obj, input)
            if size(input,2) ~= obj.indim,
				error('Input vector dimensionality does not match ARBFN dimensionality.');
            end
            
            states_expanded = repmat(input, obj.nodeCount, 1);
			product = ((states_expanded - obj.mu) ./ obj.sigma).^2;
			g = exp( -sum(product, 2) );
        end % function Rbf
		
		
        % Calculates the output of the network based on current RBFs
		function [val act_outp] = Calculate(obj, input)
			val = 0;
            if obj.nodeCount == 0,
				return;
            end
			
            g = obj.Rbf(input);
			g_expanded = repmat(g, 1, obj.outdim);
			val = sum(obj.w .* g_expanded);
            
             outpMax = max(g);
             act_outp = zeros(length(g),1);%this should be of same lengt as w
             for i = 1 : length(g)
                act_outp(i) = abs( g(i) / outpMax );
             end
		end % function Calculate
        
        
        % Adds a RBF to the network at the speicified location and with a
        % specified width.
        function obj = Add(obj, center, width, weight)
            if size(center,2) ~= obj.indim || size(width,2) ~= obj.indim,
                error('Input vector dimensionality does not match ARBFN dimensionality.');
            end
            if size(weight,2) ~= obj.outdim,
                error('Output vector dimensionality does not match ARBFN dimensionality.');
            end
            
            obj = obj.IncreasePn();
            
            obj.mu = [obj.mu;center];
            obj.sigma = [obj.sigma;width];
            obj.w = [obj.w;weight]; % NB: Note that this may also change the outputs of neighbouring RBFs.
            
            obj.nodeCount = obj.nodeCount + 1;
        end % function Add
        
        
        % The MRAN algorithm
        % input - the input vector (x_n)
        % diff - difference between calculated and actual output e_n = (y_n - ARBFN(x_n))
        function obj = Train(obj, input, diff)
            
            % Update RMS window
            obj.e_i = [diff ; obj.e_i(1:end-1)];
            obj.M_window_current = min(obj.M_window, obj.M_window_current+1);
            
            % Calculate Euclidean distance between measurement and nearest center
            [mu_nr, epsilon_calculated] = obj.FindNearestCenter(input);
            
            if obj.nodeCount == 0, % Always add in this special case
                obj = obj.Add(input, repmat(obj.sigma_0, 1, obj.indim), diff);
            else
                e_rmsn = sqrt( sum( obj.e_i(1:obj.M_window_current).^2 ) / obj.M_window_current );
                %disp(['DBG: e_rmsn = ' num2str(e_rmsn) ', ep_calc = ' num2str(epsilon_calculated) ', diff = ' num2str(diff)]);
                
                % Check if a new RBF should be added
                %disp(['e_rmsn = ' num2str(e_rmsn)]);
                if (epsilon_calculated > obj.epsilon_n) && (abs(diff) > obj.e_min) && (e_rmsn > obj.e_minl),
                    % Add new RBF
                    %disp('DBG: Adding new RBF');
                    obj = obj.Add(input, repmat(obj.Kappa*epsilon_calculated, 1, obj.indim), diff);
                else
                    % Adjust existinRg BFN by EKF algorithm
                    mu_vector = reshape(obj.mu',numel(obj.mu),1);
                    sigma_vector = reshape(obj.sigma',numel(obj.sigma),1);
                    
                    w_kalman = [ obj.w_0;
                                 obj.w;
                                 mu_vector;
                                 sigma_vector ];
                    
                    g = obj.Rbf(input);
                    g_expanded = repmat(g,1,obj.indim);
                    g_expanded = reshape(g_expanded',numel(g_expanded),1);
                    
                    w_expanded = repmat(obj.w,1,obj.indim);
                    w_expanded = reshape(w_expanded',numel(w_expanded),1);
                             
                    % Gradient vector ( Nabla_w f(x) )
                    a_n = [ 1;
                        g;
                        2 * g_expanded .* w_expanded ./ sigma_vector.^2 .* (repmat(input',obj.nodeCount,1) - mu_vector);
                        2 * g_expanded .* w_expanded ./ sigma_vector.^3 .* (repmat(input',obj.nodeCount,1) - mu_vector).^2]; % Istf. euclidean dist ||x-mu||^2 som i orig.alg. benytter jeg elementvis (x-mu).^2 siden jeg har elliptisk basisfunksjon (flerdimensjonal sigma)
                        
                    % Kalman gain vector
                    k_n = (obj.R_n + a_n' * obj.P_n * a_n) \ (obj.P_n * a_n); % (R_n + a_n'*P_n*a_n)^(-1) * (P_n*a_n) % (Slash denotes matlab inverse)
                    
                    % Kalman error covariance matrix
                    obj.P_n = (eye(size(obj.P_n)) - k_n * a_n') * obj.P_n + obj.Q*eye(size(obj.P_n));
                    
                    % Update parameters according to Kalman filter
                    w_kalman = w_kalman + diff * k_n;
                    w_index = 1;
                    obj.w_0 = w_kalman(w_index);
                    w_index = w_index + 1;
                    obj.w = w_kalman(w_index:w_index + (obj.nodeCount-1));
                    w_index = w_index + numel(obj.w);
                    obj.mu = reshape( w_kalman(w_index:w_index + (numel(obj.mu)-1))', obj.indim, obj.nodeCount )';
                    w_index = w_index + numel(obj.mu);
                    obj.sigma = reshape( w_kalman(w_index:w_index + (numel(obj.sigma)-1))', obj.indim, obj.nodeCount )';
                    
                    
                    % TODO: Prune tree
                end
            end
            
        end % function Train
        
        
        % Add new elements to the Kalman filter error covariance matrix
        % when new RBF nodes are added.
        function obj = IncreasePn(obj)
            if obj.nodeCount == 0,
                obj.P_n = obj.P_0; % Add error covariance for w_0
            end
            
            obj.P_n = [obj.P_n ...
                zeros(size(obj.P_n,1), obj.outdim + 2*obj.indim); ...
                zeros(obj.outdim + 2*obj.indim, size(obj.P_n,2)) obj.P_0*eye(obj.outdim + 2*obj.indim)];
        end % function IncreasePn
        
        
        % Finds the nearest center vector using the Euclidean distance
        function [mu_nr, dist] = FindNearestCenter(obj, x_n)
            
            if size(x_n,2) ~= obj.indim,
                error('Input vector dimensionality does not match ARBFN dimensionality.');
            end
            
            if obj.nodeCount == 0,
                mu_nr = 0;
                dist = Inf;
                return;
            end
            
            states_expanded = repmat(x_n, obj.nodeCount, 1);
            
            diff_total = sqrt( sum((states_expanded - obj.mu).^2, 2) );
            [dist,mu_nr] = min(diff_total);
            
        end % function FindNearestCenter
		
        
        
        
         function obj = remove_neuron(obj, index, n_inp)
                      
            %disp(['in remove neuron, index is  ' num2str(neuron) ]);
            
            obj.nodeCount = obj.nodeCount - 1;
            
            %disp(['in pruneO, nodeCount is now ' num2str(obj.nodeCount) ]);
            weight = obj.w;
            length_w = length(weight);
            new_weight = weight([1:index-1 index+1:length_w]');
            obj.w = new_weight(:);
                
            %disp(['in pruneO, weights are now ' num2str(length(obj.w)) ]);
               
            if n_inp == 1
                centers = obj.mu;
                length_mu = length(centers);
                centers_new = centers([1:index-1 index+1:length_mu]',1);
                %centers_new2 = centers([1:index-1 index+1:length_mu]',2);
                obj.mu = centers_new;
           elseif n_inp ==2
                centers = obj.mu;
                length_mu = length(centers);
                centers_new1 = centers([1:index-1 index+1:length_mu]',1);
                centers_new2 = centers([1:index-1 index+1:length_mu]',2);
                obj.mu = [centers_new1 centers_new2];                   
           end
                
           %disp(['in pruneO, number centers are now ' num2str(length(obj.mu)) ]);
                
           if n_inp == 1
                width = obj.sigma;
                length_sigma = length(width);
                sigma_new = width([1:index-1 index+1:length_sigma]');
                obj.sigma = sigma_new;
           elseif n_inp ==2
                width = obj.sigma;
                length_sigma = length(width);
                sigma_new1 = width([1:index-1 index+1:length_sigma]',1);
                sigma_new2 = width([1:index-1 index+1:length_sigma]',2);
                obj.sigma = [sigma_new1 sigma_new2];                
           end
           %disp(['after removing- centers:  ' num2str(length(obj.mu)) ' weights: ' num2str(length(obj.w)) ' sigmas: ' num2str(length(obj.sigma)) ]);
           %disp(['removed neuron ' num2str(neuron) ' and nodeCount is ' num2str(obj.nodeCount)]);
        
        end %remove_neuron
        
        
        function [obj act_outp pruned_units prun_Out] = pruneO(obj, act_outp, outp_th, n_inp) 
            %prunes based on size of hidden neuron output
            %Returns the obj, normalized activation function output, a list of
            %which units that have been pruned and the current pruning
            %threshold.
            %Takes as input the a list of previous normalized activation function output,
            %pruning percent and number of NN inputs

            iterations = length(obj.mu);
            
            if outp_th ~= 0
                m = 0;
                small_output = zeros(iterations,1);
                prun_Out = outp_th * max(act_outp); 

                %finds the neurons that are to be pruned away
                for i = 1 : iterations
                    if act_outp(i) < prun_Out
                        m = m+1;
                        small_output(m) = i;
                    end
                end 

                if m == 0
                    pruned_units = 0;
                else
                    pruned_units = zeros(m,1);

                    disp(['in pruneO ' num2str(m) ' units to be removed']);
                    %prunes the network
                    for k = 1 : m 
                        if k == 1
                            index = small_output(k);
                            pruned_units(k) = small_output(k);
                        end
                        if k > 1
                            index = small_output(k) - (k - 1);
                            pruned_units(k) = small_output(k);
                        end               

                        %disp(['in pruneO, index is ' num2str(index) ]);
                        obj = obj.remove_neuron(index, n_inp);    
                        act_len = length(act_outp);
                        act_tmp = act_outp([1:index-1 index+1:act_len]');
                        act_outp = act_tmp;
                        %disp(['in pruneO, activation func are now ' num2str(length(obj.sigma)) ]);

                    end %pruning
                end  %if else
            else
                pruned_units = 0;
                prun_Out = 0;               
            end
                
            %disp(['done pruneO, number centers are now ' num2str(length(obj.mu)) ]);
        end %function pruneO
        
        
        function [obj n_prunTimes pruned_units min_size act_outp] = pruneW(obj, weight_th, n_inp, n_prunTimes, act_outp)
            %prunes based on the magnitude of the weights.
            %Returns the obj, the number the NN has been pruned, a list of
            %which units that have been pruned and the current pruning
            %threshold.
            %Takes as input the pruning percent, number of NN inputs,
            %number of times the NN has been pruned before.
            
            %disp(['in pruneW ']);
            
            weight = obj.w;
            
            if weight_th ~= 0
                iterations = length(weight);
                j = 0;
                euc_dist = norm(weight,2);
                if n_prunTimes <=3 %to avoid the threshold to grow to infinity
                    min_size = (euc_dist * weight_th * n_prunTimes); %threshold for size of weights to be removed 
                else
                    min_size = (euc_dist * weight_th *1);
                    %disp(['in RBFN.m, this network has been pruned more than three times']);
                end
                small_weights = zeros(iterations,1);

                %finds the weights that are to be pruned away
                for i = 1 : iterations
                    if abs(weight(i)) < min_size
                        j = j + 1;
                        small_weights(j) = i;                        
                    end
                end

                if j == 0
                    pruned_units = 0;
                else
                    pruned_units = zeros(j,1);
                end
                disp(['in pruneW, ' num2str(j) ' weights/neurons to be removed' ]);

                for k = 1 : j
                    if k == 1
                        index = small_weights(k);
                        pruned_units(k) = small_weights(k); 
                    end
                    if k > 1
                        index = small_weights(k) - (k - 1);
                        pruned_units(k) = small_weights(k);
                    end                     
                    obj = obj.remove_neuron(index, n_inp);

                   %if the network also is pruned by node output
                    act_len = length(act_outp);
                    act_tmp = act_outp([1:index-1 index+1:act_len]');
                    act_outp = act_tmp;
                end

                if j > 0
                    n_prunTimes = n_prunTimes + 1;
                end
                
            else
                %disp(['this network can not be pruned anymore']);
                pruned_units = 0;
                min_size = 0;
                %act_outp = act_outp;
                %n_prunTimes = n_prunTimes;
            end
      
           % disp(['in pruneW, done' ]);
        end %function pruneW
        
	end % methods

end % classdef ARBFN
